# Licensed under the MIT License
# https://github.com/craigahobbs/ollama-chat/blob/main/LICENSE

"""
ollama-chat command-line script main module
"""

import argparse
import threading
import webbrowser

import waitress

from .app import OllamaChatApplication


def main(argv=None):
    """
    ollama-chat command-line script main entry point
    """

    # Command line arguments
    parser = argparse.ArgumentParser(prog='ollama-chat')
    parser.add_argument('-c', metavar='FILE', dest='config', default='ollama-chat.json',
                        help='the configuration file (default is "ollama-chat.json")')
    parser.add_argument('-p', metavar='N', dest='port', type=int, default=8080,
                        help='the application port (default is 8080)')
    parser.add_argument('-n', dest='no_browser', action='store_true',
                        help="don't open a web browser")
    parser.add_argument('-q', dest='quiet', action='store_true',
                        help="don't display access logging")
    args = parser.parse_args(args=argv)

    # Construct the URL
    host = '127.0.0.1'
    url = f'http://{host}:{args.port}/'

    # Launch the web browser on a thread so the WSGI application can startup first
    if not args.no_browser:
        webbrowser_thread = threading.Thread(target=webbrowser.open, args=(url,))
        webbrowser_thread.daemon = True
        webbrowser_thread.start()

    # Create the WSGI application
    wsgiapp = OllamaChatApplication(args.config)

    # Wrap the WSGI application and the start_response function so we can log status and environ
    def wsgiapp_wrap(environ, start_response):
        def log_start_response(status, response_headers):
            if not args.quiet:
                print(f'ollama-chat: {status[0:3]} {environ["REQUEST_METHOD"]} {environ["PATH_INFO"]} {environ["QUERY_STRING"]}')
            return start_response(status, response_headers)
        return wsgiapp(environ, log_start_response)

    # Host the application
    if not args.quiet:
        print(f'ollama-chat: Serving at {url} ...')
    waitress.serve(wsgiapp_wrap, port=args.port)



# Licensed under the MIT License
# https://github.com/craigahobbs/ollama-chat/blob/main/LICENSE

"""
The ollama chat manager
"""

import threading

import ollama


class OllamaChat():
    """
    The ollama chat manager class
    """

    __slots__ = ('app', 'conversation_id', 'stop')


    def __init__(self, app, conversation_id):
        self.app = app
        self.conversation_id = conversation_id
        self.stop = False

        # Start the chat thread
        chat_thread = threading.Thread(target=OllamaChat.chat_thread_fn, args=(self,))
        chat_thread.daemon = True
        chat_thread.start()


    @staticmethod
    def chat_thread_fn(chat):
        try:
            # Create the Ollama messages from the conversation
            messages = []
            with chat.app.config() as config:
                conversation = config_conversation(config, chat.conversation_id)
                model = conversation['model']
                for exchange in conversation['exchanges']:
                    messages.append({'role': 'user', 'content': exchange['user']})
                    if exchange['model'] != '':
                        messages.append({'role': 'assistant', 'content': exchange['model']})

            # Start the chat
            stream = ollama.chat(model=model, messages=messages, stream=True)

            # Stream the chat response
            for chunk in stream:
                # If stopped, return immediately. The chat is deleted by the stopper.
                if chat.stop:
                    stream.close()
                    break

                # Update the conversation
                with chat.app.config() as config:
                    conversation = config_conversation(config, chat.conversation_id)
                    exchange = conversation['exchanges'][-1]
                    exchange['model'] += chunk['message']['content']

        except Exception as exc: # pylint: disable=broad-exception-caught
            # Communicate the error
            with chat.app.config() as config:
                conversation = config_conversation(config, chat.conversation_id)
                exchange = conversation['exchanges'][-1]
                exchange['model'] += f'\n**ERROR:** {exc}'

        # Save the conversation
        with chat.app.config(save=True):
            # Delete the application's chat entry
            if chat.conversation_id in chat.app.chats:
                del chat.app.chats[chat.conversation_id]


def config_conversation(config, id_):
    """
    Helper to find a conversation by ID
    """
    return next((conv for conv in config['conversations'] if conv['id'] == id_), None)
